"""
    domonic.html
    ====================================
    Generate HTML using python 3
"""
import copy
from domonic.javascript import URL
from domonic.dom import Element, Document


def render(inp, outp=''):
    """render

    Renders the input to string or to a file.

    Args:
        inp (obj): A domonic tag. For example div()
        outp (str): An optional output filename

    Returns:
        str: A HTML rendered string
    """
    if outp != '':
        with open(outp, "w+") as f:
            f.write(str(inp))
    return str(inp)


class TemplateError(IndexError):
    def __init__(self, error, message="Templating error: "):
        self.error = error
        self.hint = ""
        print(self.error)
        if str(self.error) == "list index out of range":
            self.hint = "MISSING UNDERSCORE ON AN ATTRIBUTE"
        self.message = message + self.hint
        super().__init__(self.message)


class tag(object):
    """ The class from which all html tags extend """

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

        try:
            self.content = ''.join([each.__str__() for each in args])
            self.attributes = ''.join([''' %s="%s"''' % (key.split('_', 1)[1], value) for key, value in kwargs.items()])
        except IndexError as e:
            raise TemplateError(e)
        # except Exception as e:
            # print(e)

    @property
    def content(self):  # TODO - test
        return ''.join([each.__str__() for each in self.args])

    @content.setter
    def content(self, ignore):
        self.__content = ''.join([each.__str__() for each in self.args])
        return

    @property
    def attributes(self):
        try:
            return ''.join([''' %s="%s"''' % (key.split('_', 1)[1], value) for key, value in self.kwargs.items()])
        except IndexError as e:
            raise TemplateError(e)
        # except Exception as e:
            # print(e)

    @attributes.setter
    def attributes(self, ignore):
        try:
            self.__attributes = ''.join([''' %s="%s"''' % (key.split('_', 1)[1], value) for key, value in self.kwargs.items()])
        except IndexError as e:
            raise TemplateError(e)
        # except Exception as e:
            # print(e)

    def __str__(self):
        return f"<{self.name}{self.attributes}>{self.content}</{self.name}>"

    def __mul__(self, other):
        """
        requires you to render yourself i.e.
        cells = cell()*10
        print(''.join([str(c) for c in cells]))
        """
        reproducer = []
        for i in range(other):
            reproducer.append(copy.deepcopy(self))
        return reproducer

    def __rmul__(self, other):
        """
        requires you to render yourself i.e.
        cells = cell()*10
        print(''.join([str(c) for c in cells]))
        """
        reproducer = []
        for i in range(other):
            reproducer.append(copy.deepcopy(self))
        return reproducer

    def __truediv__(self, other):
        """ use to render clones without having to parse commas yourself """
        reproducer = []
        for i in range(other):
            reproducer.append(str(self))
        return ''.join(reproducer)

    def __rtruediv__(self, other):
        """ use to render clones without having to parse commas yourself """
        reproducer = []
        for i in range(other):
            reproducer.append(str(self))
        return ''.join(reproducer)

    def __div__(self, other):
        """
        useful for prototyping as renders. to retain objects use multiply
        """
        reproducer = []
        for i in range(other):
            reproducer.append(str(self))
        return ''.join(reproducer)

    def __rdiv__(self, other):
        """
        useful for prototyping as renders. to retain objects use multiply
        """
        reproducer = []
        for i in range(other):
            reproducer.append(str(self))
        return ''.join(reproducer)

    def __or__(self, other):
        """ return self unless other is something """
        if other is not False:
            return other
        return self

    def __iadd__(self, item):
        """ adds an item to the nodes of children """
        self.args = self.args + (item,)
        return self

    def __isub__(self, item):
        """ removes an item from the list of children """
        replace_args = list(self.args)
        replace_args.remove(item)
        self.args = tuple(replace_args)
        return self

    def __getitem__(self, index):
        return self.args[index]

    def __rshift__(self, item):
        try:
            for key in item.keys():
                self.kwargs[key] = item[key]
            return self
        except Exception as e:
            print(e)
            raise ValueError

    def __getattr__(self, attr):
        """
        allows dot notation for reading attributes
        *credit to the peeps on discord/python for this one*
        """
        kwargs = super().__getattribute__('kwargs')
        # print(attr)

        if attr in kwargs:
            return kwargs[attr]

        retry = "_" + attr
        if retry in kwargs:
            return kwargs[retry]

        retry = attr[1:len(attr)]
        if retry in kwargs:
            return kwargs[retry]

        raise AttributeError

    # def __repr__(self):
    #     return f"<{self.name}{self.attributes}>{self.content}</{self.name}>"

    # def __setitem__(self,key,value):
        # self.args[key] = value
        # print(self.args[key])

    # def __enter__(self):
    # def __exit__(self ,type, value, traceback):
    # def __dir__/__getattr__/__getattribute)) - .... dot notation for attributes


class closed_tag(tag):
    def __str__(self):
        return f"<{self.name}{self.attributes} />"


def tag_init(self, *args, **kwargs):
    tag.__init__(self, *args, **kwargs)
    Element.__init__(self, *args, **kwargs)


html = type('html', (tag, Document), {'name': 'html', '__init__': tag_init})
body = type('body', (tag, Element), {'name': 'body', '__init__': tag_init})
head = type('head', (tag, Element), {'name': 'head', '__init__': tag_init})
script = type('script', (tag, Element), {'name': 'script', '__init__': tag_init})
style = type('style', (tag, Element), {'name': 'style', '__init__': tag_init})
h1 = type('h1', (tag, Element), {'name': 'h1', '__init__': tag_init})
h2 = type('h2', (tag, Element), {'name': 'h2', '__init__': tag_init})
h3 = type('h3', (tag, Element), {'name': 'h3', '__init__': tag_init})
h4 = type('h4', (tag, Element), {'name': 'h4', '__init__': tag_init})
h5 = type('h5', (tag, Element), {'name': 'h5', '__init__': tag_init})
h6 = type('h6', (tag, Element), {'name': 'h6', '__init__': tag_init})
p = type('p', (tag, Element), {'name': 'p', '__init__': tag_init})
i = type('i', (tag, Element), {'name': 'i', '__init__': tag_init})
b = type('b', (tag, Element), {'name': 'b', '__init__': tag_init})

portal = type('portal', (tag, Element), {'name': 'portal', '__init__': tag_init})

def Atag(self, *args, **kwargs):
    # print('Atag: ', args, kwargs)
    tag.__init__(self, *args, **kwargs)
    Element.__init__(self, *args, **kwargs)
    # TODO - fix BUG. this stops having no href on a tags
    if kwargs.get('_href', None) is not None:
        URL.__init__(self, url=kwargs['_href'])
    else:
        URL.__init__(self, *args, **kwargs)


def __update__(self, *args, **kwargs):
    # print('__update__: ', args, kwargs)
    URL.__update__(self)

    # TODO - fix BUG. this stops having no href on a tags
    self.kwargs['_href'] = self.href
    tag.__init__(self, *self.args, **self.kwargs)


a = type('a', (tag, Element, URL), {'name': 'a', '__init__': Atag, '__update__': __update__})

ul = type('ul', (tag, Element), {'name': 'ul', '__init__': tag_init})
ol = type('ol', (tag, Element), {'name': 'ol', '__init__': tag_init})
li = type('li', (tag, Element), {'name': 'li', '__init__': tag_init})
hr = type('hr', (tag, Element), {'name': 'hr', '__init__': tag_init})
div = type('div', (tag, Element), {'name': 'div', '__init__': tag_init})
span = type('span', (tag, Element), {'name': 'span', '__init__': tag_init})
strong = type('strong', (tag, Element), {'name': 'strong', '__init__': tag_init})
blockquote = type('blockquote', (tag, Element), {'name': 'blockquote', '__init__': tag_init})
table = type('table', (tag, Element), {'name': 'table', '__init__': tag_init})
tr = type('tr', (tag, Element), {'name': 'tr', '__init__': tag_init})
td = type('td', (tag, Element), {'name': 'td', '__init__': tag_init})
title = type('title', (tag, Element), {'name': 'title', '__init__': tag_init})
meta = type('meta', (tag, Element), {'name': 'meta', '__init__': tag_init})
form = type('form', (tag, Element), {'name': 'form', '__init__': tag_init})
label = type('label', (tag, Element), {'name': 'label', '__init__': tag_init})
submit = type('submit', (tag, Element), {'name': 'submit', '__init__': tag_init})
title = type('title', (tag, Element), {'name': 'title', '__init__': tag_init})
noscript = type('noscript', (tag, Element), {'name': 'noscript', '__init__': tag_init})
section = type('section', (tag, Element), {'name': 'section', '__init__': tag_init})
nav = type('nav', (tag, Element), {'name': 'nav', '__init__': tag_init})
article = type('article', (tag, Element), {'name': 'article', '__init__': tag_init})
aside = type('aside', (tag, Element), {'name': 'aside', '__init__': tag_init})
hgroup = type('hgroup', (tag, Element), {'name': 'hgroup', '__init__': tag_init})
address = type('address', (tag, Element), {'name': 'address', '__init__': tag_init})
pre = type('pre', (tag, Element), {'name': 'pre', '__init__': tag_init})
dl = type('dl', (tag, Element), {'name': 'dl', '__init__': tag_init})
dt = type('dt', (tag, Element), {'name': 'dt', '__init__': tag_init})
dd = type('dd', (tag, Element), {'name': 'dd', '__init__': tag_init})
figure = type('figure', (tag, Element), {'name': 'figure', '__init__': tag_init})
figcaption = type('figcaption', (tag, Element), {'name': 'figcaption', '__init__': tag_init})
em = type('em', (tag, Element), {'name': 'em', '__init__': tag_init})
small = type('small', (tag, Element), {'name': 'small', '__init__': tag_init})
s = type('s', (tag, Element), {'name': 's', '__init__': tag_init})
cite = type('cite', (tag, Element), {'name': 'cite', '__init__': tag_init})
q = type('q', (tag, Element), {'name': 'q', '__init__': tag_init})
dfn = type('dfn', (tag, Element), {'name': 'dfn', '__init__': tag_init})
abbr = type('abbr', (tag, Element), {'name': 'abbr', '__init__': tag_init})
code = type('code', (tag, Element), {'name': 'code', '__init__': tag_init})
var = type('var', (tag, Element), {'name': 'var', '__init__': tag_init})
samp = type('samp', (tag, Element), {'name': 'samp', '__init__': tag_init})
kbd = type('kbd', (tag, Element), {'name': 'kbd', '__init__': tag_init})
sub = type('sub', (tag, Element), {'name': 'sub', '__init__': tag_init})
sup = type('sup', (tag, Element), {'name': 'sup', '__init__': tag_init})
u = type('u', (tag, Element), {'name': 'u', '__init__': tag_init})
mark = type('mark', (tag, Element), {'name': 'mark', '__init__': tag_init})
ruby = type('ruby', (tag, Element), {'name': 'ruby', '__init__': tag_init})
rt = type('rt', (tag, Element), {'name': 'rt', '__init__': tag_init})
rp = type('rp', (tag, Element), {'name': 'rp', '__init__': tag_init})
bdi = type('bdi', (tag, Element), {'name': 'bdi', '__init__': tag_init})
bdo = type('bdo', (tag, Element), {'name': 'bdo', '__init__': tag_init})
span = type('span', (tag, Element), {'name': 'span', '__init__': tag_init})
ins = type('ins', (tag, Element), {'name': 'ins', '__init__': tag_init})
iframe = type('iframe', (tag, Element), {'name': 'iframe', '__init__': tag_init})
video = type('video', (tag, Element), {'name': 'video', '__init__': tag_init})
audio = type('audio', (tag, Element), {'name': 'audio', '__init__': tag_init})
canvas = type('canvas', (tag, Element), {'name': 'canvas', '__init__': tag_init})
caption = type('caption', (tag, Element), {'name': 'caption', '__init__': tag_init})
colgroup = type('colgroup', (tag, Element), {'name': 'colgroup', '__init__': tag_init})
tbody = type('tbody', (tag, Element), {'name': 'tbody', '__init__': tag_init})
thead = type('thead', (tag, Element), {'name': 'thead', '__init__': tag_init})
tfoot = type('tfoot', (tag, Element), {'name': 'tfoot', '__init__': tag_init})
th = type('th', (tag, Element), {'name': 'th', '__init__': tag_init})
fieldset = type('fieldset', (tag, Element), {'name': 'fieldset', '__init__': tag_init})
legend = type('legend', (tag, Element), {'name': 'legend', '__init__': tag_init})
button = type('button', (tag, Element), {'name': 'button', '__init__': tag_init})
select = type('select', (tag, Element), {'name': 'select', '__init__': tag_init})
datalist = type('datalist', (tag, Element), {'name': 'datalist', '__init__': tag_init})
optgroup = type('optgroup', (tag, Element), {'name': 'optgroup', '__init__': tag_init})
option = type('option', (tag, Element), {'name': 'option', '__init__': tag_init})
textarea = type('textarea', (tag, Element), {'name': 'textarea', '__init__': tag_init})
output = type('output', (tag, Element), {'name': 'output', '__init__': tag_init})  # ?----------
progress = type('progress', (tag, Element), {'name': 'progress', '__init__': tag_init})
meter = type('meter', (tag, Element), {'name': 'meter', '__init__': tag_init})
details = type('details', (tag, Element), {'name': 'details', '__init__': tag_init})
summary = type('summary', (tag, Element), {'name': 'summary', '__init__': tag_init})
menu = type('menu', (tag, Element), {'name': 'menu', '__init__': tag_init})
font = type('font', (tag, Element), {'name': 'font', '__init__': tag_init})
header = type('header', (tag, Element), {'name': 'header', '__init__': tag_init})
footer = type('footer', (tag, Element), {'name': 'footer', '__init__': tag_init})
# map_ = type('map_', (tag,), {'name': 'map_', '__init__': tag_init})
# object_ = type('object_', (tag,), {'name': 'object_', '__init__': tag_init})
# del_ = type('del_', (tag,), {'name': 'del_', '__init__': tag_init})
# time_ = type('time_', (tag,), {'name': 'time_', '__init__': tag_init})

base = type('base', (closed_tag, Element), {'name': 'base', '__init__': tag_init})
link = type('link', (closed_tag, Element), {'name': 'link', '__init__': tag_init})
meta = type('meta', (closed_tag, Element), {'name': 'meta', '__init__': tag_init})
hr = type('hr', (closed_tag, Element), {'name': 'hr', '__init__': tag_init})
br = type('br', (closed_tag, Element), {'name': 'br', '__init__': tag_init})
wbr = type('wbr', (closed_tag, Element), {'name': 'wbr', '__init__': tag_init})
img = type('img', (closed_tag, Element), {'name': 'img', '__init__': tag_init})
embed = type('embed', (closed_tag, Element), {'name': 'embed', '__init__': tag_init})
param = type('param', (closed_tag, Element), {'name': 'param', '__init__': tag_init})
source = type('source', (closed_tag, Element), {'name': 'source', '__init__': tag_init})
track = type('track', (closed_tag, Element), {'name': 'track', '__init__': tag_init})
area = type('area', (closed_tag, Element), {'name': 'area', '__init__': tag_init})
col = type('col', (closed_tag, Element), {'name': 'col', '__init__': tag_init})
input = type('input', (closed_tag, Element), {'name': 'input', '__init__': tag_init})
keygen = type('keygen', (closed_tag, Element), {'name': 'keygen', '__init__': tag_init})
command = type('command', (closed_tag, Element), {'name': 'command', '__init__': tag_init})

main = type('command', (tag, Element), {'name': 'main', '__init__': tag_init})  # TODO - y was this missing?


class doctype():
    """doctype

    Returns:
        str: <!DOCTYPE html>
    """
    def __str__(self):
        return "<!DOCTYPE html>"


class comment():
    """comment

    Args:
        content (str): Message to be rendered inside the comment tag

    Returns:
        str: "<!-- {self.content} -->
    """
    def __init__(self, content=""):
        self.content = content

    def __str__(self):
        return f"<!-- {self.content} -->"



"""
    domonic.x3d
    ====================================
    Generate x3d with python 3

"""

from domonic.html import tag, closed_tag
from domonic.dom import Node, ParentNode


def x3dom_init(self, *args, **kwargs):
    tag.__init__(self, *args, **kwargs)
    Node.__init__(self, *args, **kwargs)


x3d = X3D = type('x3d', (tag, Node, ParentNode), {'name': 'x3d', '__init__': x3dom_init})
scene = Scene = type('scene', (tag, Node, ParentNode), {'name': 'scene', '__init__': x3dom_init})
material = Material = type('material', (tag, Node, ParentNode), {'name': 'material', '__init__': x3dom_init})
appearance = Appearance = type('appearance', (tag, Node, ParentNode), {'name': 'appearance', '__init__': x3dom_init})
sphere = Sphere = type('sphere', (tag, Node, ParentNode), {'name': 'sphere', '__init__': x3dom_init})
shape = Shape = type('shape', (tag, Node, ParentNode), {'name': 'shape', '__init__': x3dom_init})
transform = Transform = type('transform', (tag, Node, ParentNode), {'name': 'transform', '__init__': x3dom_init})
timeSensor = TimeSensor = type('timeSensor', (tag, Node, ParentNode), {'name': 'timeSensor', '__init__': x3dom_init})
inline = Inline = type('inline', (tag, Node, ParentNode), {'name': 'inline', '__init__': x3dom_init})
box = Box = type('box', (tag, Node, ParentNode), {'name': 'box', '__init__': x3dom_init})
plane = Plane = type('plane', (tag, Node, ParentNode), {'name': 'plane', '__init__': x3dom_init})

# TODO - go through examples to find which are usually lower. i.e before they allowed mised cases
positionInterpolator = PositionInterpolator = type('PositionInterpolator', (tag, Node, ParentNode), {'name': 'PositionInterpolator', '__init__': x3dom_init})
route = Route = type('Route', (tag, Node, ParentNode), {'name': 'Route', '__init__': x3dom_init})
anchor = Anchor = type('Anchor', (tag, Node, ParentNode), {'name': 'Anchor', '__init__': x3dom_init})
arc2D = Arc2D = type('Arc2D', (tag, Node, ParentNode), {'name': 'Arc2D', '__init__': x3dom_init})
arcClose2D = ArcClose2D = type('ArcClose2D', (tag, Node, ParentNode), {'name': 'ArcClose2D', '__init__': x3dom_init})
audioClip = AudioClip = type('AudioClip', (tag, Node, ParentNode), {'name': 'AudioClip', '__init__': x3dom_init})
background = Background = type('Background', (tag, Node, ParentNode), {'name': 'Background', '__init__': x3dom_init})
ballJoint = BallJoint = type('BallJoint', (tag, Node, ParentNode), {'name': 'BallJoint', '__init__': x3dom_init})
billboard = Billboard = type('Billboard', (tag, Node, ParentNode), {'name': 'Billboard', '__init__': x3dom_init})
binaryGeometry = BinaryGeometry = type('BinaryGeometry', (tag, Node, ParentNode), {'name': 'BinaryGeometry', '__init__': x3dom_init})
blendedVolumeStyle = BlendedVolumeStyle = type('BlendedVolumeStyle', (tag, Node, ParentNode), {'name': 'BlendedVolumeStyle', '__init__': x3dom_init})
blendMode = BlendMode = type('BlendMode', (tag, Node, ParentNode), {'name': 'BlendMode', '__init__': x3dom_init})
block = Block = type('Block', (tag, Node, ParentNode), {'name': 'Block', '__init__': x3dom_init})
boundaryEnhancementVolumeStyle = BoundaryEnhancementVolumeStyle = type('BoundaryEnhancementVolumeStyle', (tag, Node, ParentNode), {'name': 'BoundaryEnhancementVolumeStyle', '__init__': x3dom_init})
bufferAccessor = BufferAccessor = type('BufferAccessor', (tag, Node, ParentNode), {'name': 'BufferAccessor', '__init__': x3dom_init})
bufferGeometry = BufferGeometry = type('BufferGeometry', (tag, Node, ParentNode), {'name': 'BufferGeometry', '__init__': x3dom_init})
bufferView = BufferView = type('BufferView', (tag, Node, ParentNode), {'name': 'BufferView', '__init__': x3dom_init})
cADAssembly = CADAssembly = type('CADAssembly', (tag, Node, ParentNode), {'name': 'CADAssembly', '__init__': x3dom_init})
cADFace = CADFace = type('CADFace', (tag, Node, ParentNode), {'name': 'CADFace', '__init__': x3dom_init})
cADLayer = CADLayer = type('CADLayer', (tag, Node, ParentNode), {'name': 'CADLayer', '__init__': x3dom_init})
cADPart = CADPart = type('CADPart', (tag, Node, ParentNode), {'name': 'CADPart', '__init__': x3dom_init})
cartoonVolumeStyle = CartoonVolumeStyle = type('CartoonVolumeStyle', (tag, Node, ParentNode), {'name': 'CartoonVolumeStyle', '__init__': x3dom_init})
circle2D = Circle2D = type('Circle2D', (tag, Node, ParentNode), {'name': 'Circle2D', '__init__': x3dom_init})
clipPlane = ClipPlane = type('ClipPlane', (tag, Node, ParentNode), {'name': 'ClipPlane', '__init__': x3dom_init})
collidableShape = CollidableShape = type('CollidableShape', (tag, Node, ParentNode), {'name': 'CollidableShape', '__init__': x3dom_init})
collision = Collision = type('Collision', (tag, Node, ParentNode), {'name': 'Collision', '__init__': x3dom_init})
collisionCollection = CollisionCollection = type('CollisionCollection', (tag, Node, ParentNode), {'name': 'CollisionCollection', '__init__': x3dom_init})
collisionSensor = CollisionSensor = type('CollisionSensor', (tag, Node, ParentNode), {'name': 'CollisionSensor', '__init__': x3dom_init})
color = Color = type('Color', (tag, Node, ParentNode), {'name': 'Color', '__init__': x3dom_init})
colorChaser = ColorChaser = type('ColorChaser', (tag, Node, ParentNode), {'name': 'ColorChaser', '__init__': x3dom_init})
colorDamper = ColorDamper = type('ColorDamper', (tag, Node, ParentNode), {'name': 'ColorDamper', '__init__': x3dom_init})
colorInterpolator = ColorInterpolator = type('ColorInterpolator', (tag, Node, ParentNode), {'name': 'ColorInterpolator', '__init__': x3dom_init})
colorMaskMode = ColorMaskMode = type('ColorMaskMode', (tag, Node, ParentNode), {'name': 'ColorMaskMode', '__init__': x3dom_init})
colorRGBA = ColorRGBA = type('ColorRGBA', (tag, Node, ParentNode), {'name': 'ColorRGBA', '__init__': x3dom_init})
commonSurfaceShader = CommonSurfaceShader = type('CommonSurfaceShader', (tag, Node, ParentNode), {'name': 'CommonSurfaceShader', '__init__': x3dom_init})
composedCubeMapTexture = ComposedCubeMapTexture = type('ComposedCubeMapTexture', (tag, Node, ParentNode), {'name': 'ComposedCubeMapTexture', '__init__': x3dom_init})
composedShader = ComposedShader = type('ComposedShader', (tag, Node, ParentNode), {'name': 'ComposedShader', '__init__': x3dom_init})
composedTexture3D = ComposedTexture3D = type('ComposedTexture3D', (tag, Node, ParentNode), {'name': 'ComposedTexture3D', '__init__': x3dom_init})
composedVolumeStyle = ComposedVolumeStyle = type('ComposedVolumeStyle', (tag, Node, ParentNode), {'name': 'ComposedVolumeStyle', '__init__': x3dom_init})
cone = Cone = type('Cone', (tag, Node, ParentNode), {'name': 'Cone', '__init__': x3dom_init})
coordinate = Coordinate = type('Coordinate', (tag, Node, ParentNode), {'name': 'Coordinate', '__init__': x3dom_init})
coordinateDamper = CoordinateDamper = type('CoordinateDamper', (tag, Node, ParentNode), {'name': 'CoordinateDamper', '__init__': x3dom_init})
coordinateDouble = CoordinateDouble = type('CoordinateDouble', (tag, Node, ParentNode), {'name': 'CoordinateDouble', '__init__': x3dom_init})
coordinateInterpolator = CoordinateInterpolator = type('CoordinateInterpolator', (tag, Node, ParentNode), {'name': 'CoordinateInterpolator', '__init__': x3dom_init})
cylinder = Cylinder = type('Cylinder', (tag, Node, ParentNode), {'name': 'Cylinder', '__init__': x3dom_init})
cylinderSensor = CylinderSensor = type('CylinderSensor', (tag, Node, ParentNode), {'name': 'CylinderSensor', '__init__': x3dom_init})
depthMode = DepthMode = type('DepthMode', (tag, Node, ParentNode), {'name': 'DepthMode', '__init__': x3dom_init})
directionalLight = DirectionalLight = type('DirectionalLight', (tag, Node, ParentNode), {'name': 'DirectionalLight', '__init__': x3dom_init})
dish = Dish = type('Dish', (tag, Node, ParentNode), {'name': 'Dish', '__init__': x3dom_init})
disk2D = Disk2D = type('Disk2D', (tag, Node, ParentNode), {'name': 'Disk2D', '__init__': x3dom_init})
doubleAxisHingeJoint = DoubleAxisHingeJoint = type('DoubleAxisHingeJoint', (tag, Node, ParentNode), {'name': 'DoubleAxisHingeJoint', '__init__': x3dom_init})
dynamicLOD = DynamicLOD = type('DynamicLOD', (tag, Node, ParentNode), {'name': 'DynamicLOD', '__init__': x3dom_init})
edgeEnhancementVolumeStyle = EdgeEnhancementVolumeStyle = type('EdgeEnhancementVolumeStyle', (tag, Node, ParentNode), {'name': 'EdgeEnhancementVolumeStyle', '__init__': x3dom_init})
elevationGrid = ElevationGrid = type('ElevationGrid', (tag, Node, ParentNode), {'name': 'ElevationGrid', '__init__': x3dom_init})
environment = Environment = type('Environment', (tag, Node, ParentNode), {'name': 'Environment', '__init__': x3dom_init})
extrusion = Extrusion = type('Extrusion', (tag, Node, ParentNode), {'name': 'Extrusion', '__init__': x3dom_init})
field = Field = type('Field', (tag, Node, ParentNode), {'name': 'Field', '__init__': x3dom_init})
floatVertexAttribute = FloatVertexAttribute = type('FloatVertexAttribute', (tag, Node, ParentNode), {'name': 'FloatVertexAttribute', '__init__': x3dom_init})
fog = Fog = type('Fog', (tag, Node, ParentNode), {'name': 'Fog', '__init__': x3dom_init})
fontStyle = FontStyle = type('FontStyle', (tag, Node, ParentNode), {'name': 'FontStyle', '__init__': x3dom_init})
generatedCubeMapTexture = GeneratedCubeMapTexture = type('GeneratedCubeMapTexture', (tag, Node, ParentNode), {'name': 'GeneratedCubeMapTexture', '__init__': x3dom_init})
geoCoordinate = GeoCoordinate = type('GeoCoordinate', (tag, Node, ParentNode), {'name': 'GeoCoordinate', '__init__': x3dom_init})
geoElevationGrid = GeoElevationGrid = type('GeoElevationGrid', (tag, Node, ParentNode), {'name': 'GeoElevationGrid', '__init__': x3dom_init})
geoLocation = GeoLocation = type('GeoLocation', (tag, Node, ParentNode), {'name': 'GeoLocation', '__init__': x3dom_init})
geoLOD = GeoLOD = type('GeoLOD', (tag, Node, ParentNode), {'name': 'GeoLOD', '__init__': x3dom_init})
geoMetadata = GeoMetadata = type('GeoMetadata', (tag, Node, ParentNode), {'name': 'GeoMetadata', '__init__': x3dom_init})
geoOrigin = GeoOrigin = type('GeoOrigin', (tag, Node, ParentNode), {'name': 'GeoOrigin', '__init__': x3dom_init})
geoPositionInterpolator = GeoPositionInterpolator = type('GeoPositionInterpolator', (tag, Node, ParentNode), {'name': 'GeoPositionInterpolator', '__init__': x3dom_init})
geoTransform = GeoTransform = type('GeoTransform', (tag, Node, ParentNode), {'name': 'GeoTransform', '__init__': x3dom_init})
geoViewpoint = GeoViewpoint = type('GeoViewpoint', (tag, Node, ParentNode), {'name': 'GeoViewpoint', '__init__': x3dom_init})
group = Group = type('Group', (tag, Node, ParentNode), {'name': 'Group', '__init__': x3dom_init})
hAnimDisplacer = HAnimDisplacer = type('HAnimDisplacer', (tag, Node, ParentNode), {'name': 'HAnimDisplacer', '__init__': x3dom_init})
hAnimHumanoid = HAnimHumanoid = type('HAnimHumanoid', (tag, Node, ParentNode), {'name': 'HAnimHumanoid', '__init__': x3dom_init})
hAnimJoint = HAnimJoint = type('HAnimJoint', (tag, Node, ParentNode), {'name': 'HAnimJoint', '__init__': x3dom_init})
hAnimSegment = HAnimSegment = type('HAnimSegment', (tag, Node, ParentNode), {'name': 'HAnimSegment', '__init__': x3dom_init})
hAnimSite = HAnimSite = type('HAnimSite', (tag, Node, ParentNode), {'name': 'HAnimSite', '__init__': x3dom_init})
imageTexture = ImageTexture = type('ImageTexture', (tag, Node, ParentNode), {'name': 'ImageTexture', '__init__': x3dom_init})
imageTexture3D = ImageTexture3D = type('ImageTexture3D', (tag, Node, ParentNode), {'name': 'ImageTexture3D', '__init__': x3dom_init})
imageTextureAtlas = ImageTextureAtlas = type('ImageTextureAtlas', (tag, Node, ParentNode), {'name': 'ImageTextureAtlas', '__init__': x3dom_init})
indexedFaceSet = IndexedFaceSet = type('IndexedFaceSet', (tag, Node, ParentNode), {'name': 'IndexedFaceSet', '__init__': x3dom_init})
indexedLineSet = IndexedLineSet = type('IndexedLineSet', (tag, Node, ParentNode), {'name': 'IndexedLineSet', '__init__': x3dom_init})
indexedQuadSet = IndexedQuadSet = type('IndexedQuadSet', (tag, Node, ParentNode), {'name': 'IndexedQuadSet', '__init__': x3dom_init})
indexedTriangleSet = IndexedTriangleSet = type('IndexedTriangleSet', (tag, Node, ParentNode), {'name': 'IndexedTriangleSet', '__init__': x3dom_init})
indexedTriangleStripSet = IndexedTriangleStripSet = type('IndexedTriangleStripSet', (tag, Node, ParentNode), {'name': 'IndexedTriangleStripSet', '__init__': x3dom_init})
isoSurfaceVolumeData = IsoSurfaceVolumeData = type('IsoSurfaceVolumeData', (tag, Node, ParentNode), {'name': 'IsoSurfaceVolumeData', '__init__': x3dom_init})
lineProperties = LineProperties = type('LineProperties', (tag, Node, ParentNode), {'name': 'LineProperties', '__init__': x3dom_init})
lineSet = LineSet = type('LineSet', (tag, Node, ParentNode), {'name': 'LineSet', '__init__': x3dom_init})
lOD = LOD = type('LOD', (tag, Node, ParentNode), {'name': 'LOD', '__init__': x3dom_init})
matrixTextureTransform = MatrixTextureTransform = type('MatrixTextureTransform', (tag, Node, ParentNode), {'name': 'MatrixTextureTransform', '__init__': x3dom_init})
matrixTransform = MatrixTransform = type('MatrixTransform', (tag, Node, ParentNode), {'name': 'MatrixTransform', '__init__': x3dom_init})
mesh = Mesh = type('Mesh', (tag, Node, ParentNode), {'name': 'Mesh', '__init__': x3dom_init})
metadataBoolean = MetadataBoolean = type('MetadataBoolean', (tag, Node, ParentNode), {'name': 'MetadataBoolean', '__init__': x3dom_init})
metadataDouble = MetadataDouble = type('MetadataDouble', (tag, Node, ParentNode), {'name': 'MetadataDouble', '__init__': x3dom_init})
metadataFloat = MetadataFloat = type('MetadataFloat', (tag, Node, ParentNode), {'name': 'MetadataFloat', '__init__': x3dom_init})
metadataInteger = MetadataInteger = type('MetadataInteger', (tag, Node, ParentNode), {'name': 'MetadataInteger', '__init__': x3dom_init})
metadataSet = MetadataSet = type('MetadataSet', (tag, Node, ParentNode), {'name': 'MetadataSet', '__init__': x3dom_init})
metadataString = MetadataString = type('MetadataString', (tag, Node, ParentNode), {'name': 'MetadataString', '__init__': x3dom_init})
motorJoint = MotorJoint = type('MotorJoint', (tag, Node, ParentNode), {'name': 'MotorJoint', '__init__': x3dom_init})
movieTexture = MovieTexture = type('MovieTexture', (tag, Node, ParentNode), {'name': 'MovieTexture', '__init__': x3dom_init})
mPRPlane = MPRPlane = type('MPRPlane', (tag, Node, ParentNode), {'name': 'MPRPlane', '__init__': x3dom_init})
mPRVolumeStyle = MPRVolumeStyle = type('MPRVolumeStyle', (tag, Node, ParentNode), {'name': 'MPRVolumeStyle', '__init__': x3dom_init})
multiTexture = MultiTexture = type('MultiTexture', (tag, Node, ParentNode), {'name': 'MultiTexture', '__init__': x3dom_init})
multiTextureCoordinate = MultiTextureCoordinate = type('MultiTextureCoordinate', (tag, Node, ParentNode), {'name': 'MultiTextureCoordinate', '__init__': x3dom_init})
navigationInfo = NavigationInfo = type('NavigationInfo', (tag, Node, ParentNode), {'name': 'NavigationInfo', '__init__': x3dom_init})
normal = Normal = type('Normal', (tag, Node, ParentNode), {'name': 'Normal', '__init__': x3dom_init})
normalInterpolator = NormalInterpolator = type('NormalInterpolator', (tag, Node, ParentNode), {'name': 'NormalInterpolator', '__init__': x3dom_init})
nozzle = Nozzle = type('Nozzle', (tag, Node, ParentNode), {'name': 'Nozzle', '__init__': x3dom_init})
opacityMapVolumeStyle = OpacityMapVolumeStyle = type('OpacityMapVolumeStyle', (tag, Node, ParentNode), {'name': 'OpacityMapVolumeStyle', '__init__': x3dom_init})
orientationChaser = OrientationChaser = type('OrientationChaser', (tag, Node, ParentNode), {'name': 'OrientationChaser', '__init__': x3dom_init})
orientationDamper = OrientationDamper = type('OrientationDamper', (tag, Node, ParentNode), {'name': 'OrientationDamper', '__init__': x3dom_init})
orientationInterpolator = OrientationInterpolator = type('OrientationInterpolator', (tag, Node, ParentNode), {'name': 'OrientationInterpolator', '__init__': x3dom_init})
orthoViewpoint = OrthoViewpoint = type('OrthoViewpoint', (tag, Node, ParentNode), {'name': 'OrthoViewpoint', '__init__': x3dom_init})
param = Param = type('Param', (tag, Node, ParentNode), {'name': 'Param', '__init__': x3dom_init})
particleSet = ParticleSet = type('ParticleSet', (tag, Node, ParentNode), {'name': 'ParticleSet', '__init__': x3dom_init})
physicalEnvironmentLight = PhysicalEnvironmentLight = type('PhysicalEnvironmentLight', (tag, Node, ParentNode), {'name': 'PhysicalEnvironmentLight', '__init__': x3dom_init})
physicalMaterial = PhysicalMaterial = type('PhysicalMaterial', (tag, Node, ParentNode), {'name': 'PhysicalMaterial', '__init__': x3dom_init})
pixelTexture = PixelTexture = type('PixelTexture', (tag, Node, ParentNode), {'name': 'PixelTexture', '__init__': x3dom_init})
pixelTexture3D = PixelTexture3D = type('PixelTexture3D', (tag, Node, ParentNode), {'name': 'PixelTexture3D', '__init__': x3dom_init})
planeSensor = PlaneSensor = type('PlaneSensor', (tag, Node, ParentNode), {'name': 'PlaneSensor', '__init__': x3dom_init})
pointLight = PointLight = type('PointLight', (tag, Node, ParentNode), {'name': 'PointLight', '__init__': x3dom_init})
pointSet = PointSet = type('PointSet', (tag, Node, ParentNode), {'name': 'PointSet', '__init__': x3dom_init})
polyline2D = Polyline2D = type('Polyline2D', (tag, Node, ParentNode), {'name': 'Polyline2D', '__init__': x3dom_init})
polypoint2D = Polypoint2D = type('Polypoint2D', (tag, Node, ParentNode), {'name': 'Polypoint2D', '__init__': x3dom_init})
popGeometry = PopGeometry = type('PopGeometry', (tag, Node, ParentNode), {'name': 'PopGeometry', '__init__': x3dom_init})
popGeometryLevel = PopGeometryLevel = type('PopGeometryLevel', (tag, Node, ParentNode), {'name': 'PopGeometryLevel', '__init__': x3dom_init})
positionChaser = PositionChaser = type('PositionChaser', (tag, Node, ParentNode), {'name': 'PositionChaser', '__init__': x3dom_init})
positionChaser2D = PositionChaser2D = type('PositionChaser2D', (tag, Node, ParentNode), {'name': 'PositionChaser2D', '__init__': x3dom_init})
positionDamper = PositionDamper = type('PositionDamper', (tag, Node, ParentNode), {'name': 'PositionDamper', '__init__': x3dom_init})
positionDamper2D = PositionDamper2D = type('PositionDamper2D', (tag, Node, ParentNode), {'name': 'PositionDamper2D', '__init__': x3dom_init})
positionInterpolator = PositionInterpolator = type('PositionInterpolator', (tag, Node, ParentNode), {'name': 'PositionInterpolator', '__init__': x3dom_init})
positionInterpolator2D = PositionInterpolator2D = type('PositionInterpolator2D', (tag, Node, ParentNode), {'name': 'PositionInterpolator2D', '__init__': x3dom_init})
projectionVolumeStyle = ProjectionVolumeStyle = type('ProjectionVolumeStyle', (tag, Node, ParentNode), {'name': 'ProjectionVolumeStyle', '__init__': x3dom_init})
pyramid = Pyramid = type('Pyramid', (tag, Node, ParentNode), {'name': 'Pyramid', '__init__': x3dom_init})
quadSet = QuadSet = type('QuadSet', (tag, Node, ParentNode), {'name': 'QuadSet', '__init__': x3dom_init})
radarVolumeStyle = RadarVolumeStyle = type('RadarVolumeStyle', (tag, Node, ParentNode), {'name': 'RadarVolumeStyle', '__init__': x3dom_init})
rectangle2D = Rectangle2D = type('Rectangle2D', (tag, Node, ParentNode), {'name': 'Rectangle2D', '__init__': x3dom_init})
rectangularTorus = RectangularTorus = type('RectangularTorus', (tag, Node, ParentNode), {'name': 'RectangularTorus', '__init__': x3dom_init})
refinementTexture = RefinementTexture = type('RefinementTexture', (tag, Node, ParentNode), {'name': 'RefinementTexture', '__init__': x3dom_init})
remoteSelectionGroup = RemoteSelectionGroup = type('RemoteSelectionGroup', (tag, Node, ParentNode), {'name': 'RemoteSelectionGroup', '__init__': x3dom_init})
renderedTexture = RenderedTexture = type('RenderedTexture', (tag, Node, ParentNode), {'name': 'RenderedTexture', '__init__': x3dom_init})
rigidBody = RigidBody = type('RigidBody', (tag, Node, ParentNode), {'name': 'RigidBody', '__init__': x3dom_init})
rigidBodyCollection = RigidBodyCollection = type('RigidBodyCollection', (tag, Node, ParentNode), {'name': 'RigidBodyCollection', '__init__': x3dom_init})
scalarChaser = ScalarChaser = type('ScalarChaser', (tag, Node, ParentNode), {'name': 'ScalarChaser', '__init__': x3dom_init})
scalarDamper = ScalarDamper = type('ScalarDamper', (tag, Node, ParentNode), {'name': 'ScalarDamper', '__init__': x3dom_init})
scalarInterpolator = ScalarInterpolator = type('ScalarInterpolator', (tag, Node, ParentNode), {'name': 'ScalarInterpolator', '__init__': x3dom_init})
segmentedVolumeData = SegmentedVolumeData = type('SegmentedVolumeData', (tag, Node, ParentNode), {'name': 'SegmentedVolumeData', '__init__': x3dom_init})
shadedVolumeStyle = ShadedVolumeStyle = type('ShadedVolumeStyle', (tag, Node, ParentNode), {'name': 'ShadedVolumeStyle', '__init__': x3dom_init})
shaderPart = ShaderPart = type('ShaderPart', (tag, Node, ParentNode), {'name': 'ShaderPart', '__init__': x3dom_init})
silhouetteEnhancementVolumeStyle = SilhouetteEnhancementVolumeStyle = type('SilhouetteEnhancementVolumeStyle', (tag, Node, ParentNode), {'name': 'SilhouetteEnhancementVolumeStyle', '__init__': x3dom_init})
singleAxisHingeJoint = SingleAxisHingeJoint = type('SingleAxisHingeJoint', (tag, Node, ParentNode), {'name': 'SingleAxisHingeJoint', '__init__': x3dom_init})
sliderJoint = SliderJoint = type('SliderJoint', (tag, Node, ParentNode), {'name': 'SliderJoint', '__init__': x3dom_init})
slopedCylinder = SlopedCylinder = type('SlopedCylinder', (tag, Node, ParentNode), {'name': 'SlopedCylinder', '__init__': x3dom_init})
snout = Snout = type('Snout', (tag, Node, ParentNode), {'name': 'Snout', '__init__': x3dom_init})
solidOfRevolution = SolidOfRevolution = type('SolidOfRevolution', (tag, Node, ParentNode), {'name': 'SolidOfRevolution', '__init__': x3dom_init})
sound = Sound = type('Sound', (tag, Node, ParentNode), {'name': 'Sound', '__init__': x3dom_init})
sphereSegment = SphereSegment = type('SphereSegment', (tag, Node, ParentNode), {'name': 'SphereSegment', '__init__': x3dom_init})
sphereSensor = SphereSensor = type('SphereSensor', (tag, Node, ParentNode), {'name': 'SphereSensor', '__init__': x3dom_init})
splinePositionInterpolator = SplinePositionInterpolator = type('SplinePositionInterpolator', (tag, Node, ParentNode), {'name': 'SplinePositionInterpolator', '__init__': x3dom_init})
spotLight = SpotLight = type('SpotLight', (tag, Node, ParentNode), {'name': 'SpotLight', '__init__': x3dom_init})
staticGroup = StaticGroup = type('StaticGroup', (tag, Node, ParentNode), {'name': 'StaticGroup', '__init__': x3dom_init})
stippleVolumeStyle = StippleVolumeStyle = type('StippleVolumeStyle', (tag, Node, ParentNode), {'name': 'StippleVolumeStyle', '__init__': x3dom_init})
surfaceShaderTexture = SurfaceShaderTexture = type('SurfaceShaderTexture', (tag, Node, ParentNode), {'name': 'SurfaceShaderTexture', '__init__': x3dom_init})
switch = Switch = type('Switch', (tag, Node, ParentNode), {'name': 'Switch', '__init__': x3dom_init})
texCoordDamper2D = TexCoordDamper2D = type('TexCoordDamper2D', (tag, Node, ParentNode), {'name': 'TexCoordDamper2D', '__init__': x3dom_init})
text = Text = type('Text', (tag, Node, ParentNode), {'name': 'Text', '__init__': x3dom_init})
texture = Texture = type('Texture', (tag, Node, ParentNode), {'name': 'Texture', '__init__': x3dom_init})
textureCoordinate = TextureCoordinate = type('TextureCoordinate', (tag, Node, ParentNode), {'name': 'TextureCoordinate', '__init__': x3dom_init})
textureCoordinate3D = TextureCoordinate3D = type('TextureCoordinate3D', (tag, Node, ParentNode), {'name': 'TextureCoordinate3D', '__init__': x3dom_init})
textureCoordinateGenerator = TextureCoordinateGenerator = type('TextureCoordinateGenerator', (tag, Node, ParentNode), {'name': 'TextureCoordinateGenerator', '__init__': x3dom_init})
textureProperties = TextureProperties = type('TextureProperties', (tag, Node, ParentNode), {'name': 'TextureProperties', '__init__': x3dom_init})
textureTransform = TextureTransform = type('TextureTransform', (tag, Node, ParentNode), {'name': 'TextureTransform', '__init__': x3dom_init})
textureTransform3D = TextureTransform3D = type('TextureTransform3D', (tag, Node, ParentNode), {'name': 'TextureTransform3D', '__init__': x3dom_init})
textureTransformMatrix3D = TextureTransformMatrix3D = type('TextureTransformMatrix3D', (tag, Node, ParentNode), {'name': 'TextureTransformMatrix3D', '__init__': x3dom_init})
toneMappedVolumeStyle = ToneMappedVolumeStyle = type('ToneMappedVolumeStyle', (tag, Node, ParentNode), {'name': 'ToneMappedVolumeStyle', '__init__': x3dom_init})
torus = Torus = type('Torus', (tag, Node, ParentNode), {'name': 'Torus', '__init__': x3dom_init})
touchSensor = TouchSensor = type('TouchSensor', (tag, Node, ParentNode), {'name': 'TouchSensor', '__init__': x3dom_init})
triangleSet = TriangleSet = type('TriangleSet', (tag, Node, ParentNode), {'name': 'TriangleSet', '__init__': x3dom_init})
triangleSet2D = TriangleSet2D = type('TriangleSet2D', (tag, Node, ParentNode), {'name': 'TriangleSet2D', '__init__': x3dom_init})
twoSidedMaterial = TwoSidedMaterial = type('TwoSidedMaterial', (tag, Node, ParentNode), {'name': 'TwoSidedMaterial', '__init__': x3dom_init})
uniform = Uniform = type('Uniform', (tag, Node, ParentNode), {'name': 'Uniform', '__init__': x3dom_init})
universalJoint = UniversalJoint = type('UniversalJoint', (tag, Node, ParentNode), {'name': 'UniversalJoint', '__init__': x3dom_init})
viewfrustum = Viewfrustum = type('Viewfrustum', (tag, Node, ParentNode), {'name': 'Viewfrustum', '__init__': x3dom_init})
viewpoint = Viewpoint = type('Viewpoint', (tag, Node, ParentNode), {'name': 'Viewpoint', '__init__': x3dom_init})
volumeData = VolumeData = type('VolumeData', (tag, Node, ParentNode), {'name': 'VolumeData', '__init__': x3dom_init})
worldInfo = WorldInfo = type('WorldInfo', (tag, Node, ParentNode), {'name': 'WorldInfo', '__init__': x3dom_init})
x3DAppearanceChildNode = X3DAppearanceChildNode = type('X3DAppearanceChildNode', (tag, Node, ParentNode), {'name': 'X3DAppearanceChildNode', '__init__': x3dom_init})
x3DAppearanceNode = X3DAppearanceNode = type('X3DAppearanceNode', (tag, Node, ParentNode), {'name': 'X3DAppearanceNode', '__init__': x3dom_init})
x3DBackgroundNode = X3DBackgroundNode = type('X3DBackgroundNode', (tag, Node, ParentNode), {'name': 'X3DBackgroundNode', '__init__': x3dom_init})
x3DBinaryContainerGeometryNode = X3DBinaryContainerGeometryNode = type('X3DBinaryContainerGeometryNode', (tag, Node, ParentNode), {'name': 'X3DBinaryContainerGeometryNode', '__init__': x3dom_init})
x3DBindableNode = X3DBindableNode = type('X3DBindableNode', (tag, Node, ParentNode), {'name': 'X3DBindableNode', '__init__': x3dom_init})
x3DBoundedObject = X3DBoundedObject = type('X3DBoundedObject', (tag, Node, ParentNode), {'name': 'X3DBoundedObject', '__init__': x3dom_init})
x3DChaserNode = X3DChaserNode = type('X3DChaserNode', (tag, Node, ParentNode), {'name': 'X3DChaserNode', '__init__': x3dom_init})
x3DChildNode = X3DChildNode = type('X3DChildNode', (tag, Node, ParentNode), {'name': 'X3DChildNode', '__init__': x3dom_init})
x3DColorNode = X3DColorNode = type('X3DColorNode', (tag, Node, ParentNode), {'name': 'X3DColorNode', '__init__': x3dom_init})
x3DComposableVolumeRenderStyleNode = X3DComposableVolumeRenderStyleNode = type('X3DComposableVolumeRenderStyleNode', (tag, Node, ParentNode), {'name': 'X3DComposableVolumeRenderStyleNode', '__init__': x3dom_init})
x3DComposedGeometryNode = X3DComposedGeometryNode = type('X3DComposedGeometryNode', (tag, Node, ParentNode), {'name': 'X3DComposedGeometryNode', '__init__': x3dom_init})
x3DCoordinateNode = X3DCoordinateNode = type('X3DCoordinateNode', (tag, Node, ParentNode), {'name': 'X3DCoordinateNode', '__init__': x3dom_init})
x3DDamperNode = X3DDamperNode = type('X3DDamperNode', (tag, Node, ParentNode), {'name': 'X3DDamperNode', '__init__': x3dom_init})
x3DDragSensorNode = X3DDragSensorNode = type('X3DDragSensorNode', (tag, Node, ParentNode), {'name': 'X3DDragSensorNode', '__init__': x3dom_init})
x3DEnvironmentNode = X3DEnvironmentNode = type('X3DEnvironmentNode', (tag, Node, ParentNode), {'name': 'X3DEnvironmentNode', '__init__': x3dom_init})
x3DEnvironmentTextureNode = X3DEnvironmentTextureNode = type('X3DEnvironmentTextureNode', (tag, Node, ParentNode), {'name': 'X3DEnvironmentTextureNode', '__init__': x3dom_init})
x3DFogNode = X3DFogNode = type('X3DFogNode', (tag, Node, ParentNode), {'name': 'X3DFogNode', '__init__': x3dom_init})
x3DFollowerNode = X3DFollowerNode = type('X3DFollowerNode', (tag, Node, ParentNode), {'name': 'X3DFollowerNode', '__init__': x3dom_init})
x3DFontStyleNode = X3DFontStyleNode = type('X3DFontStyleNode', (tag, Node, ParentNode), {'name': 'X3DFontStyleNode', '__init__': x3dom_init})
x3DGeometricPropertyNode = X3DGeometricPropertyNode = type('X3DGeometricPropertyNode', (tag, Node, ParentNode), {'name': 'X3DGeometricPropertyNode', '__init__': x3dom_init})
x3DGeometryNode = X3DGeometryNode = type('X3DGeometryNode', (tag, Node, ParentNode), {'name': 'X3DGeometryNode', '__init__': x3dom_init})
x3DGroupingNode = X3DGroupingNode = type('X3DGroupingNode', (tag, Node, ParentNode), {'name': 'X3DGroupingNode', '__init__': x3dom_init})
x3DInfoNode = X3DInfoNode = type('X3DInfoNode', (tag, Node, ParentNode), {'name': 'X3DInfoNode', '__init__': x3dom_init})
x3DInterpolatorNode = X3DInterpolatorNode = type('X3DInterpolatorNode', (tag, Node, ParentNode), {'name': 'X3DInterpolatorNode', '__init__': x3dom_init})
x3DLightNode = X3DLightNode = type('X3DLightNode', (tag, Node, ParentNode), {'name': 'X3DLightNode', '__init__': x3dom_init})
x3DLODNode = X3DLODNode = type('X3DLODNode', (tag, Node, ParentNode), {'name': 'X3DLODNode', '__init__': x3dom_init})
x3DMaterialNode = X3DMaterialNode = type('X3DMaterialNode', (tag, Node, ParentNode), {'name': 'X3DMaterialNode', '__init__': x3dom_init})
x3DMetadataObject = X3DMetadataObject = type('X3DMetadataObject', (tag, Node, ParentNode), {'name': 'X3DMetadataObject', '__init__': x3dom_init})
x3DNavigationInfoNode = X3DNavigationInfoNode = type('X3DNavigationInfoNode', (tag, Node, ParentNode), {'name': 'X3DNavigationInfoNode', '__init__': x3dom_init})
x3DNBodyCollidableNode = X3DNBodyCollidableNode = type('X3DNBodyCollidableNode', (tag, Node, ParentNode), {'name': 'X3DNBodyCollidableNode', '__init__': x3dom_init})
x3DNode = X3DNode = type('X3DNode', (tag, Node, ParentNode), {'name': 'X3DNode', '__init__': x3dom_init})
x3DPlanarGeometryNode = X3DPlanarGeometryNode = type('X3DPlanarGeometryNode', (tag, Node, ParentNode), {'name': 'X3DPlanarGeometryNode', '__init__': x3dom_init})
x3DPointingDeviceSensorNode = X3DPointingDeviceSensorNode = type('X3DPointingDeviceSensorNode', (tag, Node, ParentNode), {'name': 'X3DPointingDeviceSensorNode', '__init__': x3dom_init})
x3DRigidJointNode = X3DRigidJointNode = type('X3DRigidJointNode', (tag, Node, ParentNode), {'name': 'X3DRigidJointNode', '__init__': x3dom_init})
x3DSensorNode = X3DSensorNode = type('X3DSensorNode', (tag, Node, ParentNode), {'name': 'X3DSensorNode', '__init__': x3dom_init})
x3DShaderNode = X3DShaderNode = type('X3DShaderNode', (tag, Node, ParentNode), {'name': 'X3DShaderNode', '__init__': x3dom_init})
x3DShapeNode = X3DShapeNode = type('X3DShapeNode', (tag, Node, ParentNode), {'name': 'X3DShapeNode', '__init__': x3dom_init})
x3DSoundNode = X3DSoundNode = type('X3DSoundNode', (tag, Node, ParentNode), {'name': 'X3DSoundNode', '__init__': x3dom_init})
x3DSoundSourceNode = X3DSoundSourceNode = type('X3DSoundSourceNode', (tag, Node, ParentNode), {'name': 'X3DSoundSourceNode', '__init__': x3dom_init})
x3DSpatialGeometryNode = X3DSpatialGeometryNode = type('X3DSpatialGeometryNode', (tag, Node, ParentNode), {'name': 'X3DSpatialGeometryNode', '__init__': x3dom_init})
x3DTexture3DNode = X3DTexture3DNode = type('X3DTexture3DNode', (tag, Node, ParentNode), {'name': 'X3DTexture3DNode', '__init__': x3dom_init})
x3DTextureCoordinateNode = X3DTextureCoordinateNode = type('X3DTextureCoordinateNode', (tag, Node, ParentNode), {'name': 'X3DTextureCoordinateNode', '__init__': x3dom_init})
x3DTextureNode = X3DTextureNode = type('X3DTextureNode', (tag, Node, ParentNode), {'name': 'X3DTextureNode', '__init__': x3dom_init})
x3DTextureTransformNode = X3DTextureTransformNode = type('X3DTextureTransformNode', (tag, Node, ParentNode), {'name': 'X3DTextureTransformNode', '__init__': x3dom_init})
x3DTimeDependentNode = X3DTimeDependentNode = type('X3DTimeDependentNode', (tag, Node, ParentNode), {'name': 'X3DTimeDependentNode', '__init__': x3dom_init})
x3DTouchSensorNode = X3DTouchSensorNode = type('X3DTouchSensorNode', (tag, Node, ParentNode), {'name': 'X3DTouchSensorNode', '__init__': x3dom_init})
x3DTransformNode = X3DTransformNode = type('X3DTransformNode', (tag, Node, ParentNode), {'name': 'X3DTransformNode', '__init__': x3dom_init})
x3DVertexAttributeNode = X3DVertexAttributeNode = type('X3DVertexAttributeNode', (tag, Node, ParentNode), {'name': 'X3DVertexAttributeNode', '__init__': x3dom_init})
x3DViewpointNode = X3DViewpointNode = type('X3DViewpointNode', (tag, Node, ParentNode), {'name': 'X3DViewpointNode', '__init__': x3dom_init})
x3DVolumeDataNode = X3DVolumeDataNode = type('X3DVolumeDataNode', (tag, Node, ParentNode), {'name': 'X3DVolumeDataNode', '__init__': x3dom_init})
x3DVolumeRenderStyleNode = X3DVolumeRenderStyleNode = type('X3DVolumeRenderStyleNode', (tag, Node, ParentNode), {'name': 'X3DVolumeRenderStyleNode', '__init__': x3dom_init})


"""
    domonic
    ====================================
    - Generate HTML using python 3
    - DOM-like API in python 3
    - JS-like API in python 3
    - Call Terminal commands using python 3 (this one requires a nix machine)
"""

__version__ = "0.2.17"
__license__ = 'MIT'

# from typing import *

from .html import *
# from .dom import Element
from .javascript import *
# from .terminal import *

import requests
import re

class domonic:

    JS_MASTER = "assets/js/master.js"
    CSS_STYLE = "assets/css/style.css"

    @staticmethod
    def domonify(domonic_string: str):
        """ [
            domonify attempts to turn a domonic_string string back into a python object.
        ]

        Args:
            dominc_string (str): [a string in the form div(_class="123")]

        Returns:
            a python object
        """
        return eval(domonic_string)

    @staticmethod
    def get(url: str):
        """ downloads html and converts to domonic """
        r = requests.get(url)
        # print(r.text.decode("utf-8"))
        return domonic.parse(r.content.decode("utf-8"))

    @staticmethod
    def output(domonic):
        """ outputs HTML str """
        # TODO - prettify using newlines to save installing this?
        # from html5print import HTMLBeautifier
        # render(HTMLBeautifier.beautify(render(output), 4), 'index.html')
        pass

    @staticmethod
    def parse(html: str) -> str:
        """ HTML as input and returns python """

        # NOTE - not working/finished.
        print("attempting to parse the page")

        html = ''.join(html.split('<!DOCTYPE HTML>'))
        html = ''.join(html.split('<!doctype html>'))

        # html = "<html><body>some content</body></html>"

        htmltags = ["figcaption", "blockquote", "textarea", "progress", "optgroup", "noscript", "fieldset", "datalist",
                    "colgroup", "summary", "section", "details", "command", "caption", "article", "address", "submit",
                    "strong", "source", "select", "script", "output", "option", "object", "legend", "keygen", "iframe",
                    "hgroup", "header", "footer", "figure", "canvas", "button", "video", "track", "title", "title",
                    "thead", "tfoot", "tbody", "table", "style", "small", "param", "meter", "label", "input", "embed",
                    "audio", "aside", "time", "span", "span", "samp", "ruby", "meta", "meta", "menu", "mark", "link",
                    "html", "head", "form", "font", "code", "cite", "body", "base", "area", "abbr", "wbr", "var", "sup",
                    "sub", "pre", "nav", "map", "main", "kbd", "ins", "img", "div", "dfn", "del", "col", "bdo", "bdi",
                    "ul", "tr", "th", "td", "rt", "rp", "ol", "li", "hr", "hr", "h6", "h5", "h4", "h3", "h2", "h1",
                    "em", "dt", "dl", "dd", "br", "u", "s", "q", "p", "i", "b", "a"]

        for tag in htmltags:
            # print(tag)
            reg = f"<{tag}>"
            pattern = re.compile(reg)
            html = re.sub(pattern, f'{tag}(', html)  # , flags=re.IGNORECASE )

            # second pass. atrributed
            reg = f"<{tag}"
            pattern = re.compile(reg)
            html = re.sub(pattern, f'{tag}(', html)  # , flags=re.IGNORECASE )

            reg = f"</{tag}>"
            pattern = re.compile(reg)
            html = re.sub(pattern, '),', html)  # , flags=re.IGNORECASE )

            reg = '/>'
            pattern = re.compile(reg)
            html = re.sub(pattern, '),', html)  # , flags=re.IGNORECASE )

        # close any tags that arent properly self closing
        flag = False
        increase_index = 0  # by the amount of new chars you add
        for index, char in enumerate(html):
            index = index + increase_index
            if char == "(":
                flag = True
                tag = html[index - 4] + html[index - 3] + \
                    html[index - 2] + html[index - 1]
                print(tag)

            if char == ")":
                flag = False
            if char == ">" and flag is True:
                if 'meta' in tag or 'link' in tag or 'hr' in tag:
                    # replace it for a '),'
                    html = f"{html[:index]}),{html[index+1:]}"
                    increase_index += 1
                else:
                    # replace for a ','
                    html = f'{html[:index]},{html[index+1:]}'

        # strip any comments
        cleaned = []
        for line in html.splitlines():
            if "<!" in line:
                continue
            cleaned.append(line)
        html = '\n'.join(cleaned)

        # attributes = ["alt", "integrity", "crossorigin", "charset", "lang",
        #               "class", "id", "style", "placeholder", "text", "value",
        #               "href", "width", "height", "src", "name", "content",
        #               "rel", "color", "type", "size"]

        attributes = [
            "accept", "accesskey", "action", "align", "alt", "async", "autocomplete", "autofocus",
            "autoplay", "bgcolor", "border", "charset", "checked", "cite", "class", "color",
            "cols", "colspan", "content", "contenteditable", "controls", "coords", "data", "datetime", "default",
            "defer", "dir", "dirname", "disabled", "download", "draggable", "enctype", "for", "form", "formaction",
            "headers", "height", "hidden", "high", "href", "hreflang", "id", "ismap", "kind", "label", "lang", "list",
            "loop", "low", "max", "maxlength", "media", "method", "min", "multiple", "muted", "name", "novalidate",
            "onabort", "onafterprint", "onbeforeprint", "onbeforeunload", "onblur", "oncanplay", "oncanplaythrough",
            "onchange", "onclick", "oncontextmenu", "oncopy", "oncuechange", "oncut", "ondblclick", "ondrag",
            "ondragend", "ondragenter", "ondragleave", "ondragover", "ondragstart", "ondrop", "ondurationchange",
            "onemptied", "onended", "onerror", "onfocus", "onhashchange", "oninput", "oninvalid", "onkeydown",
            "onkeypress", "onkeyup", "onload", "onloadeddata", "onloadedmetadata", "onloadstart", "onmousedown",
            "onmousemove", "onmouseout", "onmouseover", "onmouseup", "onmousewheel", "onoffline", "ononline",
            "onpagehide", "onpageshow", "onpaste", "onpause", "onplay", "onplaying", "onpopstate", "onprogress",
            "onratechange", "onreset", "onresize", "onscroll", "onsearch", "onseeked", "onseeking", "onselect",
            "onstalled", "onstorage", "onsubmit", "onsuspend", "ontimeupdate", "ontoggle", "onunload", "onvolumechange",
            "onwaiting", "onwheel", "open", "optimum", "pattern", "placeholder", "poster", "preload", "readonly",
            "rel", "required", "reversed", "rows", "rowspan", "sandbox", "scope", "selected", "shape", "size", "sizes",
            "span", "spellcheck", "src", "srcdoc", "srclang", "srcset", "start", "step", "style", "tabindex", "target",
            "title", "translate", "type", "usemap", "value", "width", "wrap", "property"]

        # put underscores on all the attributes
        for attr in attributes:
            reg = f' {attr}="'
            pattern = re.compile(reg)
            # , flags=re.IGNORECASE )
            html = re.sub(pattern, f' _{attr}="', html)

        # commas between them
        for attr in attributes:
            reg = f'" _{attr}="'
            pattern = re.compile(reg)
            # , flags=re.IGNORECASE )
            html = re.sub(pattern, f'", _{attr}="', html)

        # TODO - diff between loaded and inline
        # get the style and script tags
        htmltags = ["style", "script"]
        for tag in htmltags:
            reg = f"<{tag}>"
            pattern = re.compile(reg)
            html = re.sub(
                pattern, f'{tag}("""', html)  # , flags=re.IGNORECASE )

            # second pass. atrributed
            reg = f"<{tag}"
            pattern = re.compile(reg)
            html = re.sub(
                pattern, f'{tag}("""', html)  # , flags=re.IGNORECASE )

            reg = f"</{tag}>"
            pattern = re.compile(reg)
            html = re.sub(pattern, '"""),', html)  # , flags=re.IGNORECASE )

            reg = '/>'
            pattern = re.compile(reg)
            html = re.sub(pattern, '"""),', html)  # , flags=re.IGNORECASE )

        html = ')'.join(html.split(',)'))

        print(html)

        # return eval('print("test")')
        # return eval('Location()')
        # return eval('html()')
        return html



"""
    domonic.particles
    ====================================
    fun stuff

"""

from domonic.javascript import *


class Particle():

    def __init__(self, size=1):
        self.x=0
        self.y=0
        self.width=size
        self.height=size
        self.rotation=0
        self.vx=0
        self.vy=0
        self.damp=0.9
        self.bounce=-0.5
        self.grav=0
        self.wander=0
        self._k=0.2
        self.edges="wrap"
        self._drag=False
        self._oldx=None
        self._oldy=None
        self._turnToPath=False
        self._bounds = {}
        self.set_bounds( {'xMin':0, 'yMin':0, 'yMax': 800, 'xMax': 800} )
        self.maxSpeed = Number.MAX_VALUE

    def set_bounds( self, oBounds ):
        self._bounds['top'] = oBounds['yMin']
        self._bounds['bottom'] = oBounds['yMax']
        self._bounds['left'] = oBounds['xMin']
        self._bounds['right'] = oBounds['xMax']

    def turnToPath( self, bTurn ):
        self._turnToPath = bTurn

    def update( self ):
        if self._drag:
            self.vx = self.x - self._oldx
            self.vy = self.y - self._oldy
            self._oldx = self.x
            self._oldy = self.y

        else:
            self.vx += Math.random() * self.wander - self.wander / 2
            self.vy += Math.random() * self.wander - self.wander / 2
            self.vy += self.grav
            self.vx *= self.damp
            self.vy *= self.damp

            speed = Math.sqrt(self.vx * self.vx + self.vy * self.vy)
            if speed > self.maxSpeed:
                self.vx = self.maxSpeed * self.vx / speed
                self.vy = self.maxSpeed * self.vy / speed

            if self._turnToPath:
                self.rotation = Math.atan2(self.vy, self.vx)

            self.x += self.vx
            self.y += self.vy
            if self.edges == "wrap":
                if self.x > (self._bounds['right'] + self.width/2):
                    self.x = self._bounds['left'] - self.width/2
                elif self.x<self._bounds['left'] - self.width/2:
                    self.x = self._bounds['right'] + self.width/2

                if( self.y > self._bounds['bottom'] + self.height/2):
                    self.y = self._bounds['top'] - self.height/2
                elif (self.y<self._bounds['top'] - self.height/2):
                    self.y = self._bounds['bottom'] + self.height/2

            elif self.edges == "bounce":
                if self.x > (self._bounds['right'] - self.width/2):                   
                    self.x = self._bounds['right'] - self.width/2
                    self.vx *= self.bounce
                elif self.x<(self._bounds['left'] + self.width/2):
                    self.x = self._bounds['left'] + self.width/2
                    self.vx *= self.bounce

                if self.y >(self._bounds['bottom'] - self.height/2):                                 
                    self.y = self._bounds['bottom'] - self.height/2
                    self.vy *= self.bounce
                elif self.y<(self._bounds['top'] + self.height/2):
                    self.y = self._bounds['top'] + self.height/2
                    self.vy *= self.bounce

            # else if self.edges == "remove":
            #     if( self.x > self._bounds.right + self.width/2 || self.x<self._bounds.left - self.width/2 ||
            #        self.y > self._bounds.bottom + self.height/2 || self.y<self._bounds.top - self.height/2){
        
        if Global.isNaN(self.x):
            self.x=self.y=self.vx=self.vy=0




class Particle3D():

    def __init__(self,  size=1):
        self.x=0
        self.y=0
        self.z=0
        self.width=size
        self.height=size
        self.length=size
        self.rotation=0
        self.vx=0
        self.vy=0
        self.vz=0
        self.damp=0.9
        self.bounce=-0.5
        self.grav=0
        self.wander=0
        self._k=0.2
        self.edges="wrap"
        self._drag=False
        self._oldx=None
        self._oldy=None
        self._oldz=None
        self._turnToPath=False
        self._bounds = {}
        self.set_bounds( {'xMin':0, 'yMin':0, 'yMax': 800, 'xMax': 800} )
        self.maxSpeed = Number.MAX_VALUE

    def set_bounds( self, oBounds ):
        self._bounds['top'] = oBounds['yMin']
        self._bounds['bottom'] = oBounds['yMax']
        self._bounds['left'] = oBounds['xMin']
        self._bounds['right'] = oBounds['xMax']

    def turnToPath( self, bTurn ):
        self._turnToPath = bTurn

    def update( self ):
        if self._drag:
            self.vx = self.x - self._oldx
            self.vy = self.y - self._oldy
            self.vz = self.z - self._oldz
            self._oldx = self.x
            self._oldy = self.y
            self._oldz = self.z

        else:
            self.vx += Math.random() * self.wander - self.wander / 2
            self.vy += Math.random() * self.wander - self.wander / 2
            self.vz += Math.random() * self.wander - self.wander / 2
            self.vy += self.grav
            self.vx *= self.damp
            self.vy *= self.damp
            self.vz *= self.damp

            speed = Math.sqrt(self.vx * self.vx + self.vy * self.vy + self.vz * self.vz)
            if speed > self.maxSpeed:
                self.vx = self.maxSpeed * self.vx / speed
                self.vy = self.maxSpeed * self.vy / speed
                self.vz = self.maxSpeed * self.vz / speed

            if self._turnToPath:
                self.rotation = Math.atan2(self.vy, self.vx)

            self.x += self.vx
            self.y += self.vy
            self.z += self.vz

            if self.edges == "wrap":
                if self.x > (self._bounds['right'] + self.width/2):
                    self.x = self._bounds['left'] - self.width/2
                elif self.x<self._bounds['left'] - self.width/2:
                    self.x = self._bounds['right'] + self.width/2

                if( self.y > self._bounds['bottom'] + self.height/2):
                    self.y = self._bounds['top'] - self.height/2
                elif (self.y<self._bounds['top'] - self.height/2):
                    self.y = self._bounds['bottom'] + self.height/2

            elif self.edges == "bounce":
                if self.x > (self._bounds['right'] - self.width/2):                   
                    self.x = self._bounds['right'] - self.width/2
                    self.vx *= self.bounce
                elif self.x<(self._bounds['left'] + self.width/2):
                    self.x = self._bounds['left'] + self.width/2
                    self.vx *= self.bounce

                if self.y >(self._bounds['bottom'] - self.height/2):                                 
                    self.y = self._bounds['bottom'] - self.height/2
                    self.vy *= self.bounce
                elif self.y<(self._bounds['top'] + self.height/2):
                    self.y = self._bounds['top'] + self.height/2
                    self.vy *= self.bounce

            # else if self.edges == "remove":
            #     if( self.x > self._bounds.right + self.width/2 || self.x<self._bounds.left - self.width/2 ||
            #        self.y > self._bounds.bottom + self.height/2 || self.y<self._bounds.top - self.height/2){
        
        if Global.isNaN(self.x):
            self.x=self.y=self.vx=self.vy=0




"""
    test_domonic
    ~~~~~~~~~~~~
    - unit tests for domonic.dom

"""

import unittest
# import requests
# from mock import patch

from domonic.html import *
from domonic.style import *
from domonic.dom import *
from domonic import *


class domonicTestCase(unittest.TestCase):
    """ Tests for the dom package """

    def test_dom_Node(self):

        n = Node()
        # print(n)
        self.assertIsInstance(n, Node)
        # n.assertEqual(str(sometag), '<div id="someid">asdfasdf<div></div><div>yo</div></div>')
        # n.baseURI = 'eventual.technology'
        # n.baseURIObject = None
        # n.isConnected = True
        # n.namespaceURI = "http://www.w3.org/1999/xhtml"
        # n.nodePrincipal = None
        # n.outerText = None
        # n.ownerDocument = None
        # n.prefix = None  # 🗑️
        # n.rootNode = None

        b = Node()
        n.appendChild(b)
        self.assertEqual(True, n.hasChildNodes)

        c = Node()
        n.appendChild(c)
        self.assertEqual(c, n.lastChild)
        self.assertEqual(b, n.firstChild)
        self.assertEqual(2, n.childElementCount)
        self.assertEqual(True, b in n.childNodes)
        self.assertEqual(True, c in n.childNodes)
        self.assertEqual(None, n.localName)  # obsolete if not a tag or attribute should return none
        self.assertEqual(2, len(n.children))
        self.assertEqual(None, n.nodeValue)
        # print(n.nodeType())
        d = div("test")
        # print(type(d))
        # print(d.nodeName)
        self.assertEqual("DIV", d.nodeName)
        self.assertEqual("test", d.nodeValue)
        self.assertEqual(True, n.contains(c))

        n.insertBefore(d, c)
        self.assertEqual(True, n.children[1] == d)

        self.assertEqual(True, n.contains(c))
        n.removeChild(c)
        self.assertEqual(False, n.contains(c))

        # print( n.replaceChild(self, newChild, oldChild) )
        n2 = n.cloneNode()
        # print(len(n2.children))
        self.assertEqual(True, len(n2.children) == 2)
        self.assertEqual(False, n.children == n2.children)
        self.assertEqual(True, n.isSameNode(n))
        self.assertEqual(False, n.isSameNode(n2))
        a1 = div("hi")
        a2 = div("hi")
        self.assertEqual(True, a1.isEqualNode(a2))

        print(a1.nodeValue)
        self.assertEqual(True, a1.nodeValue == "hi")

        a1.nodeValue = "something else"
        self.assertEqual(True, a1.nodeValue == "something else")
        print(a1.nodeValue)

        a1.textContent = "something new"
        self.assertEqual(True, a1.textContent == "something new")
        print(a1.textContent)

        myobj = domonic.domonify('div(_class="mytest")')
        print('---')
        print(type(myobj))
        myobj.style.float = "left"
        print('---')
        print(myobj)
        self.assertEqual(True, str(myobj) == '<div class="mytest" style="float:left;"></div>')

        mylist = li() / 10
        myobj = domonic.domonify(domonic.parse(mylist))
        print(myobj)

        myorderedlist = ol()
        myorderedlist += str(li() / 10)
        print(myorderedlist)

        # compareDocumentPosition()
        # getRootNode()
        # isDefaultNamespace()
        # lookupNamespaceURI()
        # lookupPrefix()
        # normalize()
        # def isSupported(self): return False #  🗑
        # getUserData() 🗑️
        # setUserData() 🗑️

    def test_dom_node(self):
        sometag = div("asdfasdf", div(), div("yo"), _id="test", _thingy="test22")
        somenewdiv = div('im new')
        sometag.appendChild(somenewdiv)

        print("parent tests")
        print(somenewdiv.parentNode)
        print(type(somenewdiv.parentNode))
        print(somenewdiv.parentNode.id)
        print(somenewdiv.parentElement)
        print(somenewdiv.previousSibling)
        print(somenewdiv.previousSibling.nextSibling)

        mylist = ul(li(1), li(2), li(3))
        print(mylist[1])

        mylist = ul(li(), li(), li())
        print(*mylist)

        a1, b1, c1 = ul(li(1), li(2), li(3))
        print(a1)

        a1, b1, c1, d1, e1 = button() * 5
        print(a1, b1, c1, d1, e1)

        # print(mylist[1] != mylist[1])
        a1 = img()
        a1 >> {'_src': "http://www.someurl.com"}
        print(a1)

        a1 = button()
        a1 += "hi"
        a1 += "how"
        a1 += "are"
        a1 += "you"
        print(a1)
        a1 -= "hi"
        # print(a1)
        # print(div(_test="1", **{"_data-test": ""}))

        print(sometag.id)
        # print(sometag.style.color)  # TODO - get on style

        print(sometag._thingy)
        print(sometag.thingy)
        # print(10/sometag)

        # print('>>>>', sometag.args[0])
        # print('>>>>',sometag)
        # print('>>>>', sometag.lastChild())
        # print('>>>>', sometag.content)

        # import gc
        # import pprint
        # for r in gc.get_referents(somenewdiv):
        #     pprint.pprint(r)

        # for r in gc.get_referents(sometag):
        #     pprint.pprint(r)

    def test_dom_node_again(self):
        somebody = body("test", _class="why")  # .html("wn")
        print(somebody)

        somebody = body("test", _class="why").html("nope")
        print(somebody)

    def test_dom(self):

        # test div html and innerhtml update content
        sometag = div("asdfasdf", div(), div("yo"), _id="someid")
        self.assertEqual(sometag.tagName, 'div')
        self.assertEqual(str(sometag), '<div id="someid">asdfasdf<div></div><div>yo</div></div>')
        sometag.html('test')
        self.assertEqual(str(sometag), '<div id="someid">test</div>')
        sometag.innerHTML = 'test2'
        self.assertEqual(str(sometag), '<div id="someid">test2</div>')

        # same test on body tag
        bodytag = body("test", _class="why")
        self.assertEqual(str(bodytag), '<body class="why">test</body>')
        # print(bodytag)

        bodytag.html("bugs bunny")
        self.assertEqual(str(bodytag), '<body class="why">bugs bunny</body>')
        # print('THIS:',bodytag)

        # sometag.innerText()
        print(sometag.getAttribute('_id'))
        self.assertEqual(sometag.getAttribute('_id'), 'someid')
        print(sometag.getAttribute('id'))
        self.assertEqual(sometag.getAttribute('_id'), 'someid')

        mydiv = div("I like cake", div(_class='myclass').html(div("1"), div("2"), div("3")))
        print(mydiv)

        # print(sometag.innerText())
        # print(sometag.nodeName)
        # assert(sometag.nodeName, 'DIV') # TODO - i checked one site in chrome, was upper case. not sure if a standard?

        print(sometag.setAttribute('id', 'newid'))
        print(sometag)

        print(sometag.lastChild)
        print(sometag.hasChildNodes)
        # print('>>',sometag.textContent()) # TODO - will have a think. either strip or render tagless somehow

        sometag.removeAttribute('id')
        print(sometag)

        sometag.appendChild(footer('test'))
        print(sometag)

        print(sometag.children)
        print(sometag.firstChild)

        htmltag = html()
        print(htmltag)
        htmltag.write('sup!')
        htmltag.className = "my_cool_css"
        print(htmltag)
        print('-END-')

    def test_dom_create(self):
        print(html().documentElement)
        print(html().URL)
        somebody = document.createElement('sometag')
        print(str(somebody))
        comm = document.createComment('hi there here is a comment')
        print(comm)

        print(html().createElement('sometag'))
        # somebody = document.createElement('sometag')
        # print(str(somebody()))

    def test_dom_events(self):
        print(html().documentElement)
        print(html().URL)
        site = html()
        somebody = document.createElement('div')
        site.appendChild(somebody)
        print(site)

        def test(evt, *args, **kwargs):
            print('test ran!')
            print(evt)
            print(evt.target)

        site.addEventListener('click', test)
        somebody.addEventListener('anything', test)
        print(site.listeners)
        # site.removeEventListener('click', test)
        # print( site.listeners )

        site.dispatchEvent(Event('click'))
        somebody.dispatchEvent(Event('anything'))

        # document.getElementById("myBtn").addEventListener("click", function(){
        #   document.getElementById("demo").innerHTML = "Hello World";
        # });

    def test_dom_contains(self):
        site = html()
        somebody = document.createElement('div')
        site.appendChild(somebody)
        print(site)
        another_div = div()
        print(site.contains(somebody))
        another_div = div()
        print(site.contains(another_div))
        another_div = document.createElement('div')
        print(site.contains(another_div))
        third_div = document.createElement('div')
        another_div.appendChild(third_div)
        site.appendChild(another_div)
        print(site.contains(third_div))

    def test_dom_getElementById(self):
        dom1 = html(div(div(div(div(div(div(div(div("asdfasdf", div(), div("yo"), _id="test")))))))))
        result = dom1.getElementById('test')
        print('--')
        print(result)
        print('--')
        pass

    def test_dom_remove(self):
        dom1 = html(div(div(div(div(div(div(div(div("asdfasdf", div(), div("yo"), _id="test")))))))))
        result = dom1.getElementById('test')
        print("owner:", result.ownerDocument)
        result.remove()
        print(result)
        print('--')
        print(dom1)
        print('--')
        pass

    # def test_dom_getElementByClassName(self):
    #     dom1 = html(div(div(div(div(div(div(div(div("asdfasdf", div(), div("yo"), _class="test this thing")))))))))
    #     result = dom1.getElementByClassName('thing')
    #     print('--')
    #     print(result)
    #     print('--')
    #     pass

    def test_dom_dir(self):
        dom1 = div(div(), _dir="rtl")
        print('--')
        print(dom1.dir)
        print('--')
        pass

    def test_dom_normalize(self):
        dom1 = html()
        wrapper = dom1.createElement("div")
        wrapper.appendChild(dom1.createTextNode("Part 1 "))
        wrapper.appendChild(dom1.createTextNode("Part 2 "))
        wrapper.appendChild("Part 3 ")

        print('--')
        print(dom1)
        print('--')
        print(len(wrapper.childNodes))  # 2
        wrapper.normalize()
        print(len(wrapper.childNodes))  # 1
        print(wrapper.childNodes[0].textContent)  # "Part 1 Part 2 "
        pass

    # def test_dom_Node():
        # TODO - tests all below
        # contains - probably need more recursive testing
        # replaceChild
        # anchors

    def test_dom_querySelector(self):
        dom1 = html(div(div(div(div(div(div(div( div(_id="thing"), span(_id="fun"), div("asdfasdf", div(), div("yo"), _class="test this thing")))))))))

        result = dom1.querySelector('#thing')
        # print('--')
        print("RESULT>>>>>", result)
        # print('--')

        result = dom1.querySelector('span')
        # print('--')
        print("RESULT>>>>>", result)

        result = dom1.querySelector('.test')
        # print('--')
        print("RESULT>>>>>", result)

        result = dom1.getElementsByClassName('this')
        # print('--')
        print("RESULT>>>>>", result)

        pass


if __name__ == '__main__':
    unittest.main()



"""
    domonic.JSON
    ====================================
"""

import json
import csv

from domonic.html import table, td, tr, th
# from domonic.javascript import Array


def return_json(func):
    """ decorate any function to return json instead of a python obj """
    def JSON_decorator(*args, **kwargs):
        return json.dumps(func(*args, **kwargs))
    return JSON_decorator


class JSON(object):
    """ A class containing JSON utils """

    def __init__(self):
        pass

    @staticmethod
    def parse_file(filepath: str):
        f = open(filepath)
        data = json.load(f)
        f.close()
        return data

    @staticmethod
    def parse(json_string):
        return json.loads(json_string)

    @staticmethod
    def stringify(data, filepath: str = None, **kwargs):
        if filepath is not None:
            json.dump(data, filepath, **kwargs)
            return json.dumps(data, **kwargs)
        return json.dumps(data, **kwargs)  # indent=4, sort_keys=True, default=str

    @staticmethod
    def tablify(arr):
        """tablify

        takes a json array and returns a html table
        # TODO - reverse. table to json

        Args:
            arr (list): the json array

        Returns:
            str: a html table
        """
        def _get_headings(arr, t):
            headings = []
            row = tr()
            for each in arr:
                for key in each:
                    if key not in headings:
                        headings.append(key)
                        row.appendChild(th(key))
            t.appendChild(row)
            return headings

        if type(arr) == str: arr = json.loads(arr)  # leniency. allow for a string
        if type(arr) == dict: arr = arr[next(iter(arr))]  # leniency. allow for a dict wrapping a list
        if type(arr) != list: raise ValueError  # if it aint a list by now reject it

        t = table()
        headings = _get_headings(arr, t)
        for item in arr:
            row = tr(''.join([str(td(item.get(heading, ""))) for heading in headings]))
            t.appendChild(row)
        return t

    @staticmethod
    def csvify(arr, outfile="data.csv"):
        """csvify

        takes a json array and dumps a csv file

        Args:
            arr (list): the json array
            outfile (list): the output file

        Returns:
            str: a csv file
        """
        if type(arr) == str: arr = json.loads(arr)  # leniency. allow for a string
        if type(arr) == dict: arr = arr[next(iter(arr))]  # leniency. allow for a dict wrapping a list
        if type(arr) != list: raise ValueError  # if it aint a list by now reject it

        def _get_headings(arr):
            headings = []
            for each in arr:
                for key in each:
                    if key not in headings:
                        headings.append(key)
            return headings

        with open(outfile, "w") as file:
            output = csv.writer(file)
            output.writerow(_get_headings(arr))
            for row in arr:
                output.writerow(row.values())

    @staticmethod
    def csv2json(csv_filepath, json_filepath=None):
        '''
        convert a CSV to JSON.
        '''
        items = []
        with open(csv_filepath, encoding='utf-8') as csvf:
            csvReader = csv.DictReader(csvf)
            for row in csvReader:
                items.append(row)

        if json_filepath is None: return json.dumps(items)

        with open(json_filepath, 'w', encoding='utf-8') as f:
            f.write(json.dumps(items, indent=4))
            return json.dumps(items)

    '''
    @staticmethod
    def csv2json_hugefile(arr, infile="data.csv", start_row=0):

        def _load_data(csv_fname):
            with open(csv_fname, "r", encoding="latin-1") as records:
                for row in csv.reader(records):
                    yield row

        items = iter(load_data(infile))
        headings = next(companies)

        for i in range(start_row):
            next(companies)

        for item in items:
            # TODO - streamwrite to json file.
    '''

    @staticmethod
    def flatten(b, delim="__"):
        '''
        # i.e. input = map( lambda x: JSON.flatten( x, "__" ), input )
        '''
        val = {}
        for i in b.keys():
            if isinstance(b[i], dict):
                get = JSON.flatten(b[i], delim)
                for j in get.keys():
                    val[i + delim + j] = get[j]
            else:
                val[i] = b[i]

        return val

    # def flatten(): # completely flatten. underscore by default or based on rule
    # def nest(): # completely nest. underscore by default or based on rule
    # def purify # remove all the data leaving just the data structure/schema

    @staticmethod
    def is_JSON(json: str):
        if type(json) != str: return False

        if json.startswith('{') and json.endswith('}'):
            return True

        if json.startswith('[') and json.endswith(']'):
            return True

        return False

    # def value(self, query:str):
        # pass


'''
# ideas....

# with JSON( data, 'items') as item:
    # print(item)
    # print(item.id)

# iterator = JSON( data, 'items.age', lambda i: i<30 )

# diff
# merge
# strip('key') .
# format / minify
# json2sql
# json2sqlalchemymodel . i.e. https://www.jsonutils.com/
# csvify_stream for bigger ones.
'''



"""
    test_domonic
    ~~~~~~~~~~~~
    unit tests for css
"""

import unittest
# import json
# import requests
# from mock import patch

from domonic.html import *
from domonic.CDN import *


class domonicTestCase(unittest.TestCase):

    def test_domonic_CDN(self):
        myjs = script(_src=CDN_JS.JQUERY_3_5_1)
        print(myjs)

        # css
        mycss = link(_href=CDN_CSS.MARX)
        print(mycss)
        mycss = link(_rel="stylesheet", _href=CDN_JS.JQUERY_3_5_1)
        print(mycss)

        myimg = img(_src=CDN_IMG.PLACEHOLDER(100, 100))
        print(myimg)


if __name__ == '__main__':
    unittest.main()
